from django.db import models
from django.urls import reverse
import re

import marko
from marko.block import FencedCode, Heading
from marko.ext.gfm import GFM
from marko.html_renderer import HTMLRenderer

from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter


class CustomMarkdownRenderer(HTMLRenderer):

    def render_heading(self, element: Heading) -> str:
        heading_text = self.render_children(element)
        heading_id = re.sub(r"\W+", "-", heading_text.lower()).strip("-")
        return f'<h{element.level} id="{heading_id}">{heading_text}</h{element.level}>\n'
    
    def render_fenced_code(self, element: FencedCode) -> str:
        language = element.lang or 'text'
        code = element.children[0].children.replace("\t", "    ")
        try:
            lexer = get_lexer_by_name(language)
        except Exception:
            lexer = get_lexer_by_name('text')
        formatter = HtmlFormatter()
        return highlight(code, lexer, formatter)
    

class Article(models.Model):

    title = models.CharField(max_length=255, blank=True, null=True)
    slug = models.CharField(max_length=255, blank=True, null=True)
    lang = models.CharField(max_length=20, default="en")
    date_created = models.DateTimeField(auto_now=False, auto_now_add=True)
    date_modified = models.DateTimeField(blank=True, null=True)
    content = models.TextField(blank=True, null=True)
    published = models.BooleanField(default=True)
    date_published = models.DateTimeField(blank=True, null=True)
    teaser = models.TextField(blank=True, null=True)
    illustration = models.URLField(blank=True, null=True)
    external = models.BooleanField(default=False)
    external_url = models.URLField(blank=True, null=True)
    external_publisher = models.CharField(max_length=255, blank=True, null=True)
    html = models.TextField(blank=True)
    toc = models.TextField(blank=True, null=True)

    class Meta:

        ordering = ["-date_created"]

    def save(self, *args, **kwargs):
        self.html = self.compute_html()
        self.toc = self.compute_toc()
        models.Model.save(self, *args, **kwargs)

    def compute_html(self) -> str:
        text = self.content
        if self.lang == "fr":
            text = re.sub(r"(\w) ([:\?!;»€°])", r"\1 \2", text)
            text = re.sub(r"([«°]) (\w)", r"\1 \2", text)
        converter = marko.Markdown(renderer=CustomMarkdownRenderer, extensions=[GFM])
        return converter(text)\
            .replace("&lt;iframe", "<iframe")\
            .replace("<table>", '<div class="table-wrapper"><table>')\
            .replace("</table>", "</table></div>")
    
    def compute_toc(self) -> str | None:
        toc = """<ul class="toc">"""
        count = 0
        previous_level = None
        for m in re.finditer(r"<h(\d) id=\"([\w\-]+)\">(.+)<\/h", self.html):
            count += 1
            level = int(m.group(1))
            slug = m.group(2)
            label = m.group(3)
            if previous_level is not None and level > previous_level:
                for _ in range(previous_level, level):
                    toc += "<ul>"
            elif previous_level is not None and level < previous_level:
                for _ in range(level, previous_level):
                    toc += "</ul>"
            toc += f"""<li><a href="#{slug}">{label}</a></li>"""
            previous_level = level
        if count == 0:
            return None
        return toc + "</ul>"

    def __str__(self):
        text = f"<Article {self.id}>" if self.title is None else self.title
        if self.published:
            text += " [PUBLISHED]"
        return text

    def get_absolute_url(self):
        if self.external:
            return self.external_url
        return reverse("blog:blog_item", kwargs={"slug": self.slug})

    def duration(self):
        word_count = 0
        if self.content:
            word_count = len(self.content.split(" "))
        return round(float(word_count) / 200.)
